from flask import Blueprint, request, jsonify
from nova_server.utils.status_utils import JOBS, get_all_jobs, JobStatus
from nova_server.utils.key_utils import get_key_from_request_form

status = Blueprint("status", __name__)


@status.route("/job_status", methods=["POST"])
def job_status():
    if request.method == "POST":
        request_form = request.form.to_dict()
        status_key = get_key_from_request_form(request_form)

        if status_key in JOBS.keys():
            status = JOBS[status_key].status
            return jsonify({"status": status.value})
        else:
            return jsonify({"status": JobStatus.WAITING.value})


@status.route("/job_status_all", methods=["GET"])
def job_status_all():
    return jsonify(get_all_jobs())



from flask import Blueprint, request, jsonify
from nova_server.utils import log_utils
from nova_server.utils.key_utils import get_key_from_request_form


log = Blueprint("log", __name__)


@log.route("/log", methods=["POST"])
def log_thread():
    if request.method == "POST":
        request_form = request.form.to_dict()
        log_key = get_key_from_request_form(request_form)
        if log_key in log_utils.LOGS:
            logger = log_utils.LOGS[log_key]
            path = logger.handlers[0].baseFilename
            with open(path) as f:
                f = f.readlines()
            output = ''
            for line in f:
                output += line
            return jsonify({'message': output})
        else:
            return jsonify({'message': 'No log for the given parameters found.'})



import random

def get_key_from_request_form(request_form):
    """
    Returns the logging key from the provided request form
    Args:
        request_form (dict): Request form from Nova

    Returns:

    """
    key = request_form.get("jobID", None)

    if key is None:
        key = f"local_{random.randint(0, 10 ^ 7)}"

    # id_components = [
    #     request_form.get('username', None),
    #     request_form.get('database', None),
    #     request_form.get('scheme', None),
    #     request_form.get('streamName', None),
    #     request_form.get('annotator', None),
    #     request_form.get('sessions', None).replace(";", "_"),
    # ]
    # server_key = '_'.join([x for x in id_components if x])


    # serverkey = request_form['username'] + '_' + request_form['database'] + '_' + request_form['scheme'] + '_' + \
    #     request_form['streamName'] + '_' + request_form['annotator'] + '_' + \
    #     request_form['sessions'].replace(";", "_")
    return key



import os
import shutil
import imblearn
import traceback
from importlib.machinery import SourceFileLoader

from flask import Blueprint, request, jsonify
from nova_server.utils import dataset_utils, thread_utils, status_utils, log_utils, import_utils
from nova_server.utils.status_utils import update_progress
from nova_server.utils.key_utils import get_key_from_request_form
from nova_server.utils.thread_utils import THREADS
from pathlib import Path, PureWindowsPath
from nova_utils.ssi_utils.ssi_xml_utils import Trainer
from flask import current_app

import importlib

train = Blueprint("train", __name__)
thread = Blueprint("thread", __name__)


@train.route("/train", methods=["POST"])
def train_thread():
    if request.method == "POST":
        request_form = request.form.to_dict()
        key = get_key_from_request_form(request_form)
        thread = train_model(request_form, current_app._get_current_object())
        status_utils.add_new_job(key)
        data = {"success": "true"}
        thread.start()
        THREADS[key] = thread
        return jsonify(data)


@thread_utils.ml_thread_wrapper
def train_model(request_form, app_context):
    key = get_key_from_request_form(request_form)
    logger = log_utils.get_logger_for_thread(key)
    cml_dir = app_context.config['CML_DIR']
    data_dir = app_context.config['DATA_DIR']

    try:
        status_utils.update_status(key, status_utils.JobStatus.RUNNING)
        update_progress(key, 'Initializing')

        trainer_file_path = Path(cml_dir).joinpath(PureWindowsPath(request_form["trainerFilePath"]))
        out_dir = Path(cml_dir).joinpath(
            PureWindowsPath(request_form["trainerOutputDirectory"])
        )
        trainer_name = request_form["trainerName"]

        log_conform_request = dict(request_form)
        log_conform_request['password'] = '---'

        logger.info("Action 'Train' started.")
        trainer = Trainer()

        if not trainer_file_path.is_file():
            logger.error("Trainer file not available!")
            status_utils.update_status(key, status_utils.JobStatus.ERROR)
            return None
        else:
            trainer.load_from_file(trainer_file_path)
            logger.info("Trainer successfully loaded.")


        # Load Data
        try:
            update_progress(key, 'Data loading')
            ds_iter = dataset_utils.dataset_from_request_form(request_form, data_dir)
            logger.info("Train-Data successfully loaded...")
        except ValueError as e:
            print(e)
            log_utils.remove_log_from_dict(key)
            logger.error("Not able to load the data from the database!")
            status_utils.update_status(key, status_utils.JobStatus.ERROR)
            return None

        # Load Trainer
        model_script_path = trainer_file_path.parent / trainer.model_script_path
        source = SourceFileLoader("model_script", str(model_script_path)).load_module()
        import_utils.assert_or_install_dependencies(source.REQUIREMENTS, Path(model_script_path).stem)
        model_script = source.TrainerClass(ds_iter, logger, request_form)

        # Set Options 
        logger.info("Setting options...")
        if not request_form["OptStr"] == '':
            for k, v in dict(option.split("=") for option in request_form["OptStr"].split(";")).items():
                if v in ('True', 'False'):
                    model_script.OPTIONS[k] = True if v == 'True' else False
                elif v == 'None':
                    model_script.OPTIONS[k] = None
                else:
                    model_script.OPTIONS[k] = v
                logger.info(k + '=' + v)
        logger.info("...done.")

        logger.info("Preprocessing data...")
        # TODO: generic preprocessing interface, remove request form from model interface
        model_script.preprocess()
        logger.info("...done")

        logger.info("Training model...")
        model_script.train()
        logger.info("...done")

        # Save
        logger.info('Saving...')
        out_dir.mkdir(parents=True, exist_ok=True)
        trainer.info_trained = True

        # TODO add users / sessions
        # TODO there might be a more elegant way to get classes...
        tmp = ds_iter.annos[ds_iter.roles[0] + "." + ds_iter.schemes[0]].labels
        trainer.classes = []
        if request_form["schemeType"] == "DISCRETE_POLYGON" or request_form["schemeType"] == "POLYGON":
            for entry in tmp:
                trainer.classes.append({"name": tmp[entry][0]})
        else:
            for entry in tmp:
                trainer.classes.append({"name": tmp[entry]})

        weight_path = model_script.save(out_dir / trainer_name)
        trainer.model_weights_path = os.path.basename(weight_path)
        logger.info('...weights')
        shutil.copy(model_script_path, out_dir / trainer.model_script_path)
        logger.info('...train script')
        for f in model_script.DEPENDENCIES:
            shutil.copy(trainer_file_path.parent / f, out_dir / f)
        logger.info('...dependencies')
        trainer.write_trainer_to_file(out_dir / trainer_name)
        logger.info('...trainerfile')
        logger.info("Training completed!")
        status_utils.update_status(key, status_utils.JobStatus.FINISHED)

    except Exception as e:
        logger.error('Error:' + traceback.format_exc())
        status_utils.update_status(key, status_utils.JobStatus.ERROR)


# TODO DATA BALANCING
def balance_data(request_form, x_np, y_np):
    # DATA BALANCING
    if request_form["balance"] == "over":
        print("OVERSAMPLING from {} Samples".format(x_np.shape))
        oversample = imblearn.over_sampling.SMOTE()
        x_np, y_np = oversample.fit_resample(x_np, y_np)
        print("to {} Samples".format(x_np.shape))

    if request_form["balance"] == "under":
        print("UNDERSAMPLING from {} Samples".format(x_np.shape))
        undersample = imblearn.under_sampling.RandomUnderSampler()
        x_np, y_np = undersample.fit_resample(x_np, y_np)
        print("to {} Samples".format(x_np.shape))



from flask import Flask
from nova_server.route.train import train
from nova_server.route.extract import extract
from nova_server.route.status import status
from nova_server.route.log import log
from nova_server.route.ui import ui
from nova_server.route.cancel import cancel
from nova_server.route.predict import predict
import argparse
from pathlib import Path

def create_app(template_folder):
    print("Starting nova-backend server")
    app = Flask(__name__, template_folder=template_folder)
    app.register_blueprint(train)
    app.register_blueprint(predict)
    app.register_blueprint(extract)
    app.register_blueprint(log)
    app.register_blueprint(status)
    app.register_blueprint(ui)
    app.register_blueprint(cancel)
    print("... done!")
    return app

def run_app():
    from waitress import serve

    parser = argparse.ArgumentParser(
        description="Commandline arguments to configure the nova backend server"
    )
    parser.add_argument(
        "--host", type=str, default="0.0.0.0", help="The host ip address"
    )
    parser.add_argument(
        "--port", type=int, default=8080, help="The port the server listens on"
    )

    parser.add_argument(
        "--template_folder",
        type=str,
        default="./templates",
        help="Path for the templates to load relative to this script",
    )

    parser.add_argument(
        "--cml_dir",
        type=str,
        default="./cml",
        help="Cml folder to read the training scripts from. Same as in Nova.",
    )

    parser.add_argument(
        "--data_dir",
        type=str,
        default="./data",
        help="Data folder to read the training scripts from. Same as in Nova.",
    )

    parser.add_argument(
        "--cache_dir",
        type=str,
        default="./cache",
        help="Cache folder where all large files (e.g. model weights) are cached.",
    )

    parser.add_argument(
        "--tmp_dir",
        type=str,
        default="./tmp",
        help="Folder for temporary data storage.",
    )

    # TODO: support multiple (data) directories
    args = parser.parse_args()
    app = create_app(template_folder=args.template_folder)
    app.config['CML_DIR'] = args.cml_dir
    app.config['DATA_DIR'] = args.data_dir
    app.config['CACHE_DIR'] = args.cache_dir
    app.config['TMP_DIR'] = args.tmp_dir

    Path(app.config['CACHE_DIR']).mkdir(parents=False, exist_ok=True)
    Path(app.config['TMP_DIR']).mkdir(parents=False, exist_ok=True)

    host = args.host
    port = args.port
    serve(app, host=host, port=port)


if __name__ == "__main__":
    run_app()


"""This file contains the general logic for extracting feature streams to the filesystem"""

from flask import Blueprint, request, jsonify
from nova_server.utils import (
    dataset_utils,
    thread_utils,
    status_utils,
    log_utils,
    import_utils,
)
from nova_server.utils.key_utils import get_key_from_request_form
from flask import current_app
from nova_server.utils.thread_utils import THREADS
from nova_server.utils.status_utils import update_progress
from pathlib import Path, PureWindowsPath
from nova_utils.db_utils.nova_types import DataTypes
from nova_utils.ssi_utils.ssi_xml_utils import Chain, ChainLink
from nova_utils.ssi_utils.ssi_stream_utils import Stream, Chunk, FileTypes, NPDataTypes
from importlib.machinery import SourceFileLoader
from hcai_datasets.hcai_nova_dynamic.hcai_nova_dynamic_iterable import (
    HcaiNovaDynamicIterable,
)
import ffmpegio
import numpy as np
from nova_server.utils import db_utils


extract = Blueprint("extract", __name__)


@extract.route("/extract", methods=["POST"])
def extract_thread():
    if request.method == "POST":
        request_form = request.form.to_dict()
        key = get_key_from_request_form(request_form)
        thread = extract_data(request_form, current_app._get_current_object())
        status_utils.add_new_job(key)
        data = {"success": "true"}
        thread.start()
        THREADS[key] = thread
        return jsonify(data)


@thread_utils.ml_thread_wrapper
def extract_data(request_form, app_context):

    # Initialize
    key = get_key_from_request_form(request_form)
    logger = log_utils.get_logger_for_thread(key)
    cml_dir = app_context.config["CML_DIR"]
    data_dir = app_context.config["DATA_DIR"]

    chain_file_path = Path(cml_dir).joinpath(
        PureWindowsPath(request_form["chainFilePath"])
    )

    log_conform_request = dict(request_form)
    log_conform_request["password"] = "---"

    logger.info("Action 'Extract' started.")
    chain = Chain()

    # Load chain
    # TODO: update chain file path from request form
    if not chain_file_path.is_file():
        logger.error("Chain file not available!")
        status_utils.update_status(key, status_utils.JobStatus.ERROR)
        return None
    else:
        chain.load_from_file(chain_file_path)
        logger.info("Chain successfully loaded.")

    # At the moment we support only one link in the extractor
    if len(chain.links) > 1:
        raise AssertionError(
            "Loaded chainfile consists of more than one processing step. Currently only one processing step is supported."
        )

    # Check if chain expects inputs for all role in one sample
    multi_role_input = chain.links[0].multi_role_input

    # Load data
    logger.info("Initializing data iterators...")
    try:
        update_progress(key, "Data loading")
        sessions = request_form.pop("sessions").split(";")
        roles = request_form.pop("roles").split(";")
        iterators = []
        for session in sessions:
            request_form["sessions"] = session
            if multi_role_input:
                request_form["roles"] = ";".join(roles)
                iterators.append(
                    dataset_utils.dataset_from_request_form(request_form, data_dir)
                )
            else:
                for role in roles:
                    request_form["roles"] = role
                    iterators.append(
                        dataset_utils.dataset_from_request_form(request_form, data_dir)
                    )

        logger.info("...done")
    except ValueError as e:
        print(e)
        log_utils.remove_log_from_dict(key)
        logger.error(e)
        status_utils.update_status(key, status_utils.JobStatus.ERROR)
        return None
    except FileNotFoundError as e:
        print(e)
        log_utils.remove_log_from_dict(key)
        logger.error(e)
        status_utils.update_status(key, status_utils.JobStatus.ERROR)
        return None

    # Iterate over all sessions
    ds_iter: HcaiNovaDynamicIterable
    for ds_iter in iterators:

        # Iterate over all chain links
        cl: ChainLink
        for i, cl in enumerate(chain.links):

            # Load chain link module
            model_script_path = chain_file_path.parent / cl.script
            source = SourceFileLoader(
                "ns_cl_" + model_script_path.stem, str(model_script_path)
            ).load_module()
            import_utils.assert_or_install_dependencies(
                source.REQUIREMENTS, Path(model_script_path).stem
            )
            logger.info(f"Extraction module {Path(model_script_path).name} loaded")
            extractor_class = getattr(source, cl.create)
            extractor = extractor_class(logger, log_conform_request)
            logger.info(f"Extractor {cl.create} created")

            # Set Options
            logger.info("Setting options...")
            if request_form.get("optStr"):
                for k, v in [
                    option.split("=") for option in request_form["optStr"].split(";")
                ]:
                    if v in ("True", "False"):
                        extractor.options[k] = True if v == "True" else False
                    elif v == "None":
                        extractor.options[k] = None
                    else:
                        extractor.options[k] = v
                    logger.info(k + "=" + v)
            logger.info("...done.")

            # Check if there will be more chain links to execute
            if i + 1 != len(chain.links):

                # Assert chainability
                if not extractor.chainable:
                    raise AssertionError(
                        "Extraction module does not not support further chaining but is not the last link in chain."
                    )

                # TODO implement nova data types in the server module
                # TODO process data and write tmp session files to disk if necessary
                logger.info("Extract data...")
                data = extractor.process_data(ds_iter)
                ds_iter = extractor.to_ds_iterable(data)
                logger.info("...done")

            # Last element of chain
            else:
                logger.info("Extract data...")
                data = extractor.process_data(ds_iter)
                stream_dict = extractor.to_stream(data)
                logger.info("...done")

                # Write to disk and add to database
                if not stream_dict:
                    raise ValueError("Stream data is none")
                else:
                    logger.info("Write...")

                    for stream_id, (data_type, sr, data) in stream_dict.items():

                        suffix = request_form.get("suffix", "")
                        stream_id_sfx = stream_id + "_" + suffix if suffix else stream_id
                        file_name_db = '.'.join(stream_id_sfx.split('.')[1:])
                        out_path = (
                            Path(ds_iter.nova_data_dir)
                            / ds_iter.dataset
                            / ds_iter.sessions[0]
                            / stream_id_sfx
                        )

                        logger.info(f"\t - {out_path}")

                        # SSI-Stream
                        if data_type == DataTypes.FEATURE:
                            file_ex = '.stream'
                            ftype = FileTypes.BINARY
                            sr = sr
                            dim = data.shape[-1] if len(data.shape) > 1 else 1
                            dtype = NPDataTypes(type(data.dtype).type)
                            chunks = [
                                Chunk(f=0, t=data.shape[0] / sr, b=0, n=data.shape[0])
                            ]
                            ssi_stream = Stream(
                                ftype=ftype,
                                sr=sr,
                                dim=dim,
                                byte=data.dtype.itemsize,
                                dtype=dtype,
                                data=data,
                                chunks=chunks,
                            )
                            ssi_stream.save(out_path)

                            # Audio Data
                        elif data_type == DataTypes.AUDIO:
                            file_ex = '.wav'
                            out_path = out_path.parent / (out_path.name + file_ex)
                            ffmpegio.audio.write(out_path, int(sr), np.swapaxes(np.hstack(data),0 , -1), overwrite=True)

                        # Video Data
                        elif data_type == DataTypes.VIDEO:
                            file_ex = '.mp4'
                            out_path = out_path.parent / (out_path.name + file_ex)
                            ffmpegio.video.write(out_path, int(sr), np.vstack(data), overwrite=True)
                        else:
                            raise NotImplementedError(f"{data_type} is not supported.")


                        # Add Stream info to Databases
                        # TODO: Add dimlabels from stream
                        db_utils.write_stream_info_to_db(
                             request_form = request_form,
                             file_name = file_name_db,
                             file_ext = file_ex,
                             stream_type = data_type.name.lower(),
                             is_valid = True,
                             sr = sr,
                             dimlables = []
                         )

                    logger.info("...done")

        logger.info("Extraction completed!")
        status_utils.update_status(key, status_utils.JobStatus.FINISHED)

        logger.info("Action 'Extract' finished.")
        # TODO: Start legacy ssi chain support



import sys
from pathlib import Path
import subprocess
import site

def assert_or_install_dependencies(packages, trainer_name):
    #exec_path = Path(sys.executable)
    site_package_path = (Path(site.getsitepackages()[0]) / '..' / 'nova-server-site-packages' / trainer_name).resolve()
    site_package_path.mkdir(parents=True, exist_ok=True)

    for i,pkg in enumerate(packages):
        params = []
        # split on space while also removing double spaces
        pk = [x for x in pkg.split(' ') if x]
        if len(pk) > 1:
            params.extend(pk[1:])
        name = pk[0].split('==')

        params.append("--target={}".format(site_package_path))
        adjusted_name = str(name[0]).replace('-', '_')
        dirs = [1 if x.name.startswith(adjusted_name) else 0 for x in site_package_path.iterdir() if x.is_dir()]

        if sum(dirs) > 0:
            print(f'skip installation of {site_package_path}. Package already installed')
            pass
        else:
            install_package(pk[0], params)

    sys.path.insert(0, str(site_package_path.resolve()))

def install_package(pkg, params):
    call = [sys.executable, "-m", "pip", "install", pkg, *params]
    print(call)
    return subprocess.check_call(call)


import numpy as np


def getTopXpredictions(prediction, topLabels):

    prediction_class = []

    for i in range(0, len(prediction[0])):
        prediction_class.append((i, prediction[0][i]))

    prediction_class.sort(key=lambda x: x[1], reverse=True)

    return prediction_class[:topLabels]


def preprocess(X, net):
    X = X.copy()
    X = net["preprocess_f"](X)
    return X


def postprocess(X, color_conversion, channels_first):
    X = X.copy()
    X = iutils.postprocess_images(
        X, color_coding=color_conversion, channels_first=channels_first
    )
    return X


def image(X):
    X = X.copy()
    return ivis.project(X, absmax=255.0, input_is_postive_only=True)


def bk_proj(X):
    X = ivis.clip_quantile(X, 1)
    return ivis.project(X)


def heatmap(X):
    X = ivis.gamma(X, minamp=0, gamma=0.95)
    return ivis.heatmap(X)


def heatmapgnuplot2(X):
    X = np.abs(X)
    return ivis.heatmap(X, cmap_type="gnuplot2", input_is_postive_only=True)


def heatmapCMRmap(X):
    X = np.abs(X)
    return ivis.heatmap(X, cmap_type="CMRmap", input_is_postive_only=True)


def heatmapnipy_spectral(X):
    X = np.abs(X)
    return ivis.heatmap(X, cmap_type="nipy_spectral", input_is_postive_only=True)


def heatmap_rainbow(X):
    X = np.abs(X)
    return ivis.heatmap(X, cmap_type="rainbow", input_is_postive_only=True)


def heatmap_inferno(X):
    X = np.abs(X)
    return ivis.heatmap(X, cmap_type="inferno", input_is_postive_only=True)


def heatmap_viridis(X):
    X = np.abs(X)
    return ivis.heatmap(X, cmap_type="viridis", input_is_postive_only=True)


def heatmap_gist_heat(X):
    X = np.abs(X)
    return ivis.heatmap(X, cmap_type="gist_heat", input_is_postive_only=True)


def graymap(X):
    return ivis.graymap(np.abs(X), input_is_postive_only=True)



import sys
import bson
import copy
import warnings
import numpy as np

from hcai_datasets.hcai_nova_dynamic.nova_db_handler import NovaDBHandler
from nova_utils.ssi_utils.ssi_anno_utils import Anno, SchemeType
MAX_MONGO_DB_DOC_SIZE = 16777216
database = None
scheme = None
session = None
annotator = None
roles = None


# Format of results must be: {values: [value1, value2, value3, ...], confidences: [conf1, conf2, conf3, ...], start_frame: x, ...}
def check_format(variable, logger):
    text = ""
    if type(variable) is not dict:
        text = "Result type must be a dictionary!"
    if "values" not in variable or "confidences" not in variable:
        text = 'Result type must contain "values" and "confidences" as keys.'
    if not type(variable["values"]) is list or not type(variable["confidences"]):
        text = (
            'Results["values"] and Results["confidences"] must respectively be a list.'
        )

    if text != "":
        logger.info(text)
        raise Exception(text)


def write_stream_info_to_db(
    request_form: dict,
    file_name: str,
    file_ext: str,
    stream_type: str,
    is_valid: bool,
    sr: float,
    dimlables: list = None,
):
    # TODO check if we really need to establish a new connection to the database
    db_config_dict = {
        "ip": request_form["server"].split(":")[0],
        "port": int(request_form["server"].split(":")[1]),
        "user": request_form["username"],
        "password": request_form["password"],
    }

    db_handler = NovaDBHandler(db_config_dict=db_config_dict)
    database = request_form["database"]

    db_handler.set_data_streams(
        database=database,
        file_name=file_name,
        file_ext=file_ext,
        stream_type=stream_type,
        is_valid=is_valid,
        sr=sr,
        dimlabels=dimlables,
        overwrite=True,
    )


def write_annotation_to_db(request_form, anno: Anno, logger):
    #global database, scheme, session, annotator, roles
    #check_format(results, logger)

    # TODO check if we really need to establish a new connection to the database
    # DB Config
    db_config_dict = {
        "ip": request_form["server"].split(":")[0],
        "port": int(request_form["server"].split(":")[1]),
        "user": request_form["username"],
        "password": request_form["password"],
    }

    # Database specific options
    db_handler = NovaDBHandler(db_config_dict=db_config_dict)
    database = request_form["database"]
    session = request_form["sessions"]

    # Format data correctly
    scheme_dtype_names = anno.scheme.get_dtype().base.names
    anno_data = [
        dict(zip(scheme_dtype_names, ad.item()))
        for ad
        in anno.data
    ]


    db_handler.set_annos(
        database=database,
        session=session,
        scheme=anno.scheme.name,
        annotator=anno.annotator,
        role=anno.role,
        annos=anno_data,
    )


def write_polygons_to_db(request_form, results: dict, db_handler, logger):
    mongo_scheme = db_handler.get_mongo_scheme(scheme, database)
    mongo_annotator = db_handler.get_mongo_annotator(annotator, database)
    mongo_role = db_handler.get_mongo_role(roles, database)
    mongo_session = db_handler.get_mongo_session(session, database)

    mongo_annotations = db_handler.get_annotation_docs(
        mongo_scheme,
        mongo_session,
        mongo_annotator,
        mongo_role,
        database,
        db_handler.ANNOTATION_COLLECTION,
    )

    start_frame = int(float(results["start_frame"])) - 1

    mongo_data_id = None
    data_backup_id = None
    if mongo_annotations:
        if mongo_annotations[0]["isLocked"]:
            warnings.warn(
                f"Can't overwrite locked annotation {str(mongo_annotations[0]['_id'])}"
            )
            return ""
        else:
            mongo_data_id = mongo_annotations[0]["data_id"]
            data_backup_id = mongo_annotations[0]["data_backup_id"]
    logger.info("...fetch documents...")
    # 1. Get the doc (will get merged if other docs are there are nextEntry ID's)
    main_docs = db_handler.get_data_docs_by_prop(mongo_data_id, "_id", database)
    backup_doc = db_handler.get_docs_by_prop(
        mongo_data_id, "_id", database, db_handler.ANNOTATION_DATA_COLLECTION
    )[0]
    logger.info("...fill documents...")
    # 2. Fill the doc with the Predictions
    main_docs = update_polygon_doc(
        main_docs, results["values"], results["confidences"], start_frame
    )

    # 3. Check if the doc is too large (if, separate it)
    if len(bson.BSON.encode(main_docs)) >= MAX_MONGO_DB_DOC_SIZE:
        main_docs = separate_doc(main_docs)
    if not isinstance(main_docs, list):
        main_docs = [main_docs]
    main_docs[0]["_id"] = mongo_data_id

    logger.info("...separate docs if necessary...")
    # 4 Delete old back-up (with tail-docs)
    db_handler.delete_doc_with_tail(data_backup_id, database)

    logger.info("...upload docs...")
    # 5. Update the backup ID
    backup_doc["_id"] = data_backup_id
    db_handler.insert_doc_by_prop(
        backup_doc, database, db_handler.ANNOTATION_DATA_COLLECTION
    )
    db_handler.delete_doc_by_prop(
        mongo_data_id, "_id", database, db_handler.ANNOTATION_DATA_COLLECTION
    )

    # 6. Upload the doc(s)
    for doc in main_docs:
        db_handler.insert_doc_by_prop(
            doc, database, db_handler.ANNOTATION_DATA_COLLECTION
        )


# def separate_doc(doc) -> list:
#     left = copy.deepcopy(doc)
#     left["_id"] = bson.objectid.ObjectId()
#     right = copy.deepcopy(doc)
#     right["_id"] = bson.objectid.ObjectId()
#     half_count = int(len(doc["labels"]) / 2)
#     left["labels"] = left["labels"][0:half_count]
#     right["labels"] = right["labels"][half_count:]
#
#     result_list = []
#     if len(bson.BSON.encode(left)) >= MAX_MONGO_DB_DOC_SIZE:
#         result_list += separate_doc(left)
#     else:
#         result_list.append(left)
#
#     if len(bson.BSON.encode(right)) >= MAX_MONGO_DB_DOC_SIZE:
#         result_list += separate_doc(right)
#     else:
#         result_list.append(right)
#
#     return result_list
#
#
# def update_polygon_doc(data_doc, polygons, confidences, start_frame):
#     current_frame = -1
#     for frame_id, frame in enumerate(data_doc["labels"]):
#         if frame_id >= start_frame + len(polygons):
#             return data_doc
#         if frame_id >= start_frame:
#             polygons_per_frame = polygons[frame_id - start_frame]
#             for label_id, polygons_per_label_type in enumerate(polygons_per_frame):
#                 for polygon_id, polygon in enumerate(polygons_per_label_type):
#                     label = label_id + 1
#                     points_for_db = []
#                     points = np.reshape(
#                         polygon, newshape=[int(polygon.shape[0] / 2), 2]
#                     )
#                     for point in points:
#                         points_for_db.append({"x": int(point[0]), "y": int(point[1])})
#
#                     # delete the content of the current frame, the new prediction values have to be set
#                     if current_frame != frame_id:
#                         current_frame = frame_id
#                         data_doc["labels"][frame_id]["polygons"] = []
#
#                     data_doc["labels"][frame_id]["polygons"].append(
#                         {
#                             "label": label,
#                             "confidence": round(
#                                 confidences[frame_id - start_frame][label_id][
#                                     polygon_id
#                                 ],
#                                 2,
#                             ),
#                             "points": points_for_db,
#                         }
#                     )
#
#     return data_doc



# import tensorflow as tf
# import io
# import numpy as np
# import io as inputoutput
# import base64
# import json
# import pickle
# import warnings
# import site as s
# from tensorflow.keras.models import load_model
# from PIL import Image
# from PIL import Image as pilimage
# from flask import Blueprint, Response, request
# from lime import lime_image
# from lime import lime_tabular
# from skimage.segmentation import mark_boundaries
#
# from nova_server.utils import explain_utils, img_utils
#
# # Initialization
# s.getusersitepackages()
# warnings.simplefilter("ignore")
# graph = tf.compat.v1.get_default_graph()
# explain = Blueprint("explain", __name__)
#
#
# @explain.route("/tfexplain", methods=["POST"])
# def explain_tfexplain():
#     global graph
#     data = {"success": "failed"}
#     #      # ensure an image was properly uploaded to our endpoint
#     if request.method == "POST":
#         if request.form.get("image"):
#
#             explainer = request.args.get("explainer")
#             # with graph.as_default():
#             model_path = request.form.get("model_path")
#             model = load_model(model_path)
#
#             # read the image in PIL format
#             image64 = request.form.get("image")
#             image = base64.b64decode(image64)
#             image = Image.open(io.BytesIO(image))
#             image = img_utils.prepare_image(image, target=(224, 224))
#             image = image * (1.0 / 255)
#             # img = tf.keras.preprocessing.image.img_to_array(image)
#             prediction = model.predict(image)
#             topClass = explain_utils.getTopXpredictions(prediction, 1)
#             print(topClass[0])
#             image = np.squeeze(image)
#
#             if explainer == "GRADCAM":
#                 im = ([image], None)
#                 from tf_explain.core.grad_cam import GradCAM
#
#                 exp = GradCAM()
#                 imgFinal = exp.explain(im, model, class_index=topClass[0][0])
#                 # exp.save(imgFinal, ".", "grad_cam.png")
#
#             elif explainer == "OCCLUSIONSENSITIVITY":
#                 im = ([image], None)
#                 from tf_explain.core.occlusion_sensitivity import OcclusionSensitivity
#
#                 exp = OcclusionSensitivity()
#                 imgFinal = exp.explain(
#                     im, model, class_index=topClass[0][0], patch_size=10
#                 )
#                 # exp.save(imgFinal, ".", "grad_cam.png")
#
#             elif explainer == "GRADIENTSINPUTS":
#                 im = (np.array([image]), None)
#                 from tf_explain.core.gradients_inputs import GradientsInputs
#
#                 exp = GradientsInputs()
#                 imgFinal = exp.explain(im, model, class_index=topClass[0][0])
#                 # exp.save(imgFinal, ".", "gradients_inputs.png")
#
#             elif explainer == "VANILLAGRADIENTS":
#                 im = (np.array([image]), None)
#                 from tf_explain.core.vanilla_gradients import VanillaGradients
#
#                 exp = VanillaGradients()
#                 imgFinal = exp.explain(im, model, class_index=topClass[0][0])
#                 # exp.save(imgFinal, ".", "gradients_inputs.png")
#
#             elif explainer == "SMOOTHGRAD":
#                 im = (np.array([image]), None)
#                 from tf_explain.core.smoothgrad import SmoothGrad
#
#                 exp = SmoothGrad()
#                 imgFinal = exp.explain(im, model, class_index=topClass[0][0])
#                 # exp.save(imgFinal, ".", "gradients_inputs.png")
#
#             elif explainer == "INTEGRATEDGRADIENTS":
#                 im = (np.array([image]), None)
#                 from tf_explain.core.integrated_gradients import IntegratedGradients
#
#                 exp = IntegratedGradients()
#                 imgFinal = exp.explain(im, model, class_index=topClass[0][0])
#                 # exp.save(imgFinal, ".", "gradients_inputs.png")
#
#             elif explainer == "ACTIVATIONVISUALIZATION":
#                 # need some solution to find out and submit layers name
#                 im = (np.array([image]), None)
#                 from tf_explain.core.activations import ExtractActivations
#
#                 exp = ExtractActivations()
#                 imgFinal = exp.explain(im, model, layers_name=["activation_1"])
#                 # exp.save(imgFinal, ".", "gradients_inputs.png")
#
#             img = pilimage.fromarray(imgFinal)
#             imgByteArr = inputoutput.BytesIO()
#             img.save(imgByteArr, format="JPEG")
#             imgByteArr = imgByteArr.getvalue()
#
#             img64 = base64.b64encode(imgByteArr)
#             img64_string = img64.decode("utf-8")
#
#             data["explanation"] = img64_string
#             data["prediction"] = str(topClass[0][0])
#             data["prediction_score"] = str(topClass[0][1])
#             data["success"] = "success"
#
#     return Response(json.dumps(data), mimetype="text/plain")
#
#
# @explain.route("/innvestigate", methods=["POST"])
# def explain_innvestigate():
#     global graph
#     data = {"success": "failed"}
#     # ensure an image was properly uploaded to our endpoint
#     if request.method == "POST":
#         if request.form.get("image"):
#
#             postprocess = request.args.get("postprocess")
#             explainer = request.args.get("explainer")
#             lrpalpha = float(request.args.get("lrpalpha"))
#             lrpbeta = float(request.args.get("lrpbeta"))
#
#             with graph.as_default():
#                 model_path = request.form.get("model_path")
#                 model = load_model(model_path)
#
#                 # read the image in PIL format
#                 image64 = request.form.get("image")
#                 image = base64.b64decode(image64)
#                 image = Image.open(io.BytesIO(image))
#
#                 # preprocess the image and prepare it for classification
#                 image = img_utils.prepare_image(image, target=(224, 224))
#                 image = image * (1.0 / 255)
#
#                 # print(model.summary())
#                 model_wo_sm = iutils.keras.graph.model_wo_softmax(model)
#
#                 prediction = model.predict(image)
#                 topClass = explain_utils.getTopXpredictions(prediction, 1)
#
#                 model_wo_sm = iutils.keras.graph.model_wo_softmax(model)
#
#                 analyzer = []
#
#                 if explainer == "GUIDEDBACKPROP":
#                     analyzer = innvestigate.analyzer.GuidedBackprop(model_wo_sm)
#                 elif explainer == "GRADIENT":
#                     analyzer = innvestigate.analyzer.Gradient(model_wo_sm)
#                 elif explainer == "DECONVNET":
#                     analyzer = innvestigate.analyzer.Deconvnet(model_wo_sm)
#                 elif explainer == "LRPEPSILON":
#                     analyzer = innvestigate.analyzer.LRPEpsilon(model_wo_sm)
#                 elif explainer == "LRPZ":
#                     analyzer = innvestigate.analyzer.LRPZ(model_wo_sm)
#                 elif explainer == "LRPALPHABETA":
#                     analyzer = innvestigate.analyzer.LRPAlphaBeta(
#                         model_wo_sm, alpha=lrpalpha, beta=lrpbeta
#                     )
#                 elif explainer == "DEEPTAYLOR":
#                     analyzer = innvestigate.analyzer.DeepTaylor(model_wo_sm)
#
#                 # Applying the analyzer
#                 analysis = analyzer.analyze(image)
#
#                 imgFinal = []
#
#                 if postprocess == "GRAYMAP":
#                     imgFinal = explain_utils.graymap(analysis)[0]
#                 elif postprocess == "HEATMAP":
#                     imgFinal = explain_utils.heatmap(analysis)[0]
#                 elif postprocess == "BK_PROJ":
#                     imgFinal = explain_utils.bk_proj(analysis)[0]
#                 elif postprocess == "GNUPLOT2":
#                     imgFinal = explain_utils.heatmapgnuplot2(analysis)[0]
#                 elif postprocess == "CMRMAP":
#                     imgFinal = explain_utils.heatmapCMRmap(analysis)[0]
#                 elif postprocess == "NIPY_SPECTRAL":
#                     imgFinal = explain_utils.heatmapnipy_spectral(analysis)[0]
#                 elif postprocess == "RAINBOW":
#                     imgFinal = explain_utils.heatmap_rainbow(analysis)[0]
#                 elif postprocess == "INFERNO":
#                     imgFinal = explain_utils.heatmap_inferno(analysis)[0]
#                 elif postprocess == "GIST_HEAT":
#                     imgFinal = explain_utils.heatmap_gist_heat(analysis)[0]
#                 elif postprocess == "VIRIDIS":
#                     imgFinal = explain_utils.heatmap_viridis(analysis)[0]
#
#                 imgFinal = np.uint8(imgFinal * 255)
#
#                 img = pilimage.fromarray(imgFinal)
#                 imgByteArr = inputoutput.BytesIO()
#                 img.save(imgByteArr, format="JPEG")
#                 imgByteArr = imgByteArr.getvalue()
#
#                 img64 = base64.b64encode(imgByteArr)
#                 img64_string = img64.decode("utf-8")
#                 data["explanation"] = img64_string
#                 data["prediction"] = str(topClass[0][0])
#                 data["prediction_score"] = str(topClass[0][1])
#                 data["success"] = "success"
#
#     return Response(json.dumps(data), mimetype="text/plain")
#
#
# @explain.route("/lime", methods=["POST"])
# def explain_lime():
#     global graph
#     data = {"success": "failed"}
#
#     if request.method == "POST":
#         if request.form.get("image"):
#
#             top_labels = int(request.args.get("toplabels"))
#             hide_color = str(request.args.get("hidecolor"))
#             num_samples = int(request.args.get("numsamples"))
#             positive_only = str(request.args.get("positiveonly"))
#             num_features = int(request.args.get("numfeatures"))
#             hide_rest = str(request.args.get("hiderest"))
#
#             # read the image in PIL format
#             image64 = request.form.get("image")
#             image = base64.b64decode(image64)
#             image = Image.open(io.BytesIO(image))
#
#             with graph.as_default():
#
#                 model_path = request.form.get("model_path")
#                 model = load_model(model_path)
#
#                 img = img_utils.prepare_image(image, (224, 224))
#                 img = img * (1.0 / 255)
#                 prediction = model.predict(img)
#                 explainer = lime_image.LimeImageExplainer()
#                 img = np.squeeze(img).astype("double")
#                 explanation = explainer.explain_instance(
#                     img,
#                     model.predict,
#                     top_labels=top_labels,
#                     hide_color=hide_color == "True",
#                     num_samples=num_samples,
#                 )
#
#                 top_classes = explain_utils.getTopXpredictions(prediction, top_labels)
#
#                 explanations = []
#
#                 for cl in top_classes:
#                     temp, mask = explanation.get_image_and_mask(
#                         cl[0],
#                         positive_only=positive_only == "True",
#                         num_features=num_features,
#                         hide_rest=hide_rest == "True",
#                     )
#                     img_explained = mark_boundaries(temp, mask)
#                     img = Image.fromarray(np.uint8(img_explained * 255))
#                     img_byteArr = io.BytesIO()
#                     img.save(img_byteArr, format="JPEG")
#                     img_byteArr = img_byteArr.getvalue()
#                     img64 = base64.b64encode(img_byteArr)
#                     img64_string = img64.decode("utf-8")
#
#                     explanations.append((str(cl[0]), str(cl[1]), img64_string))
#
#                 data["explanations"] = explanations
#                 data["success"] = "success"
#
#     return Response(json.dumps(data), mimetype="text/plain")
#
#
# @explain.route("/tabular", methods=["POST"])
# def explain_tabular():
#     data = {"success": "failed"}
#
#     # TODO send sample to be explained
#
#     if request.method == "POST":
#
#         if request.form:
#
#             # data_dict = ast.literal_eval(json.loads(flask.request.data))
#
#             print("try open model")
#             with open(request.form.get("model_path"), "rb") as f:
#                 model = pickle.load(f)
#
#             train_data = json.loads(request.form.get("data"))
#             dim = json.loads(request.form.get("dim"))
#             train_data = np.asarray(train_data)
#             train_data = train_data.reshape(((int)(train_data.size / dim), dim))
#             sample = json.loads(request.form.get("sample"))
#
#             num_features = int(request.args.get("numfeatures"))
#
#             explainer = lime_tabular.LimeTabularExplainer(
#                 train_data, mode="classification", discretize_continuous=True
#             )
#             exp = explainer.explain_instance(
#                 np.asarray(sample),
#                 model.predict_proba,
#                 num_features=num_features,
#                 top_labels=1,
#             )
#
#             explanation_dictionary = {}
#
#             for entry in exp.as_list():
#                 explanation_dictionary.update({entry[0]: entry[1]})
#
#             data["explanation"] = explanation_dictionary
#             data["success"] = "success"
#
#     return Response(json.dumps(data), mimetype="text/plain")



import nova_server.utils.status_utils

from flask import Blueprint, request, jsonify
from nova_server.utils import status_utils
from nova_server.utils.key_utils import get_key_from_request_form
from nova_server.utils.thread_utils import THREADS
from nova_server.utils.log_utils import LOGS


cancel = Blueprint("cancel", __name__)


@cancel.route("/cancel", methods=["POST"])
def complete_thread():
    if request.method == "POST":
        request_form = request.form.to_dict()
        key = get_key_from_request_form(request_form)

        if key in THREADS:
            thread = THREADS[key]
            thread.raise_exception()
            status_utils.update_status(key, status_utils.JobStatus.WAITING)
            if key in LOGS:
                log = LOGS[key]
                log.info("Action successfully canceled.")
            return jsonify({'success': "true"})
        else:
            if key in LOGS:
                log = LOGS[key]
                log.info("Cancel was not successful.")
            return jsonify({'success': "false"})



from datetime import datetime
from enum import Enum
from nova_server.utils.thread_utils import status_thread_wrapper
import copy

JOBS = {}


class JobStatus(Enum):
    WAITING = 0
    RUNNING = 1
    FINISHED = 2
    ERROR = 3


class Job:
    def __init__(self, job_key, interactive_url=None, log_path=None):
        self.start_time = None
        self.end_time = None
        self.progress = None
        self.status = JobStatus.WAITING
        self.job_key = job_key
        self.interactive_url = interactive_url
        self.log_path = log_path

    def serializable(self):
        s = copy.deepcopy(vars(self))
        for key in s.keys():
            s[key] = str(s[key])
        return s


@status_thread_wrapper
def add_new_job(job_key, interactive_url=None):
    job = Job(job_key, interactive_url)
    JOBS[job_key] = job

    return True


@status_thread_wrapper
def remove_job(job_key):
    try:
        del JOBS[job_key]
    except KeyError:
        print(f"Key {job_key} is not in the dictionary")



@status_thread_wrapper
def update_status(job_key, status: JobStatus):
    try:
        JOBS[job_key].status = status

        if status == status.RUNNING:
            JOBS[job_key].start_time = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
        if status == status.FINISHED or status == status.ERROR:
            JOBS[job_key].end_time = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
    except KeyError:
        print(f"Key {job_key} is not in the dictionary")



@status_thread_wrapper
def update_progress(job_key, progress: str):
    try:
        JOBS[job_key].progress = progress
    except KeyError:
        print(f"Key {job_key} is not in the dictionary")



@status_thread_wrapper
def set_log_path(job_key, log_path):
    try:
        JOBS[job_key].log_path = log_path
    except KeyError:
        print(f"Key {job_key} is not in the dictionary")



@status_thread_wrapper
def get_log_path(job_key):
    try:
        return JOBS[str(job_key)].log_path
    except KeyError:
        print(f"Key {job_key} is not in the dictionary")



@status_thread_wrapper
def get_all_jobs():
    return [job.serializable() for job in JOBS.values()]


from hcai_datasets.hcai_nova_dynamic.hcai_nova_dynamic_iterable import (
    HcaiNovaDynamicIterable,
)


def dataset_from_request_form(request_form, data_dir):
    """
    Creates a tensorflow dataset from nova dynamically
    :param request_form: the requestform that specifices the parameters of the dataset
    """
    db_config_dict = {
        "ip": request_form["server"].split(":")[0],
        "port": int(request_form["server"].split(":")[1]),
        "user": request_form["username"],
        "password": request_form["password"],
    }

    flattenSamples = False
    if request_form.get("flattenSamples") == "true":
        flattenSamples = True

    ds_iter = HcaiNovaDynamicIterable(
        # Database Config
        db_config_path=None,  # os.path.join(os.path.dirname(os.path.abspath(__file__)), 'db.cfg'),
        db_config_dict=db_config_dict,
        # Dataset Config
        dataset=request_form.get("database"),
        nova_data_dir=data_dir,
        sessions=request_form.get("sessions").split(";")
        if request_form.get("sessions")
        else None,
        roles=request_form.get("roles", "").split(";")
        if request_form.get("roles")
        else None,
        schemes=request_form.get("scheme", "").split(";")
        if request_form.get("scheme")
        else None,
        annotator=request_form.get("annotator"),
        data_streams=request_form.get("streamName", "").split(" ")
        if request_form.get("streamName")
        else None,
        # Sample Config
        frame_size=request_form.get("frameSize"),
        stride=request_form.get("stride"),
        left_context=request_form.get("leftContext"),
        right_context=request_form.get("rightContext"),
        start=request_form.get("startTime"),
        end=request_form.get("endTime"),
        # TODO: This does not work with pytorch bridge when set to true because the data field does not contain the role anymore<.
        # transformation cannot be applied. fix it!
        flatten_samples=flattenSamples,
        supervised_keys=[
            request_form.get("streamName", "").split(" ")[0],
            request_form.get("scheme", "").split(";")[0],
        ],
        # Additional Config
        clear_cache=True,
    )

    return ds_iter



"""This file contains the general logic for predicting annotations to the nova database"""
import copy
from pathlib import Path, PureWindowsPath
from nova_server.utils import db_utils
from flask import Blueprint, request, jsonify
from nova_utils.ssi_utils.ssi_xml_utils import Trainer
from importlib.machinery import SourceFileLoader
from nova_server.utils.thread_utils import THREADS
from nova_server.utils.status_utils import update_progress
from nova_server.utils.key_utils import get_key_from_request_form
from nova_server.utils import (
    thread_utils,
    status_utils,
    log_utils,
    dataset_utils,
    import_utils,
)
from hcai_datasets.hcai_nova_dynamic.hcai_nova_dynamic_iterable import (
    HcaiNovaDynamicIterable,
)
from nova_utils.interfaces.server_module import Trainer as iTrainer
from flask import current_app

predict = Blueprint("predict", __name__)


@predict.route("/predict", methods=["POST"])
def predict_thread():
    if request.method == "POST":
        request_form = request.form.to_dict()
        key = get_key_from_request_form(request_form)
        thread = predict_data(request_form, current_app._get_current_object())
        status_utils.add_new_job(key)
        data = {"success": "true"}
        thread.start()
        THREADS[key] = thread
        return jsonify(data)


@thread_utils.ml_thread_wrapper
def predict_data(request_form, app_context):
    key = get_key_from_request_form(request_form)
    logger = log_utils.get_logger_for_thread(key)
    cml_dir = app_context.config["CML_DIR"]
    data_dir = app_context.config["DATA_DIR"]

    log_conform_request = dict(request_form)
    log_conform_request["password"] = "---"

    logger.info("Action 'Predict' started.")
    status_utils.update_status(key, status_utils.JobStatus.RUNNING)
    trainer_file_path = Path(cml_dir).joinpath(
        PureWindowsPath(request_form["trainerFilePath"])
    )
    trainer = Trainer()

    if not trainer_file_path.is_file():
        logger.error("Trainer file not available!")
        status_utils.update_status(key, status_utils.JobStatus.ERROR)
        return None
    else:
        trainer.load_from_file(trainer_file_path)
        logger.info("Trainer successfully loaded.")

    if not trainer.model_script_path:
        logger.error('Trainer has no attribute "script" in model tag.')
        status_utils.update_status(key, status_utils.JobStatus.ERROR)
        return None

    # Load data
    try:
        update_progress(key, "Data loading")
        sessions = request_form.pop("sessions").split(";")
        roles = request_form.pop("roles").split(";")
        iterators = []
        for session in sessions:
            request_form["sessions"] = session
            if trainer.model_multi_role_input:
                request_form["roles"] = ";".join(roles)
                iterators.append(
                    dataset_utils.dataset_from_request_form(request_form, data_dir)
                )
            else:
                for role in roles:
                    request_form["roles"] = role
                    iterators.append(
                        dataset_utils.dataset_from_request_form(request_form, data_dir)
                    )

        logger.info("Data iterators initialized.")
    except ValueError as e:
        print(e)
        log_utils.remove_log_from_dict(key)
        logger.error("Not able to load the data from the database!")
        status_utils.update_status(key, status_utils.JobStatus.ERROR)
        return None

    # Load Trainer
    model_script_path = (trainer_file_path.parent / PureWindowsPath(trainer.model_script_path)).resolve()
    source = SourceFileLoader(
        "ns_tr_" + model_script_path.stem, str(model_script_path)
    ).load_module()
    import_utils.assert_or_install_dependencies(
        source.REQUIREMENTS, Path(model_script_path).stem
    )
    logger.info(f"Trainer module {Path(model_script_path).name} loaded")
    trainer_class = getattr(source, trainer.model_create)
    predictor = trainer_class(logger, log_conform_request)
    logger.info(f"Model {trainer.model_create} created")
    # model_script = source.TrainerClass(ds_iter, logger, request_form)

    # If the module implements the Trainer interface load weights
    if isinstance(predictor, iTrainer):
        # Load Model
        model_weight_path = (
                trainer_file_path.parent / trainer.model_weights_path
        )
        logger.info(f"Loading weights from {model_weight_path}")
        predictor.load(model_weight_path)
        logger.info("Model loaded.")

    # Iterate over all sessions
    ds_iter: HcaiNovaDynamicIterable
    for ds_iter in iterators:
        # TODO: Remove prior creation of separate iterators to reduce redundancy
        ss_ds_iter = ds_iter.to_single_session_iterator()

        logger.info("Predict data...")
        data = predictor.process_data(ss_ds_iter)
        annos = predictor.to_anno(data)
        logger.info("...done")

        logger.info("Saving predictions to database...")

        # TODO: Refactor to not use request form in upload
        request_form_copy = copy.copy(request_form)
        assert len(ss_ds_iter.sessions) == 1
        request_form_copy['sessions'] = ss_ds_iter.sessions[0]

        for anno in annos:
            db_utils.write_annotation_to_db(request_form_copy, anno, logger)
        logger.info("...done")

    logger.info("Prediction completed!")
    status_utils.update_status(key, status_utils.JobStatus.FINISHED)

'''Keep for later reference to implement polygons'''
    # model_script.ds_iter = ds_iter
    # model_script.request_form["sessions"] = session
    # model_script.request_form["roles"] = role
    #
    # logger.info("Execute preprocessing.")
    # model_script.preprocess()
    # logger.info("Preprocessing done.")
    #
    # logger.info("Execute prediction.")
    # model_script.predict()
    # logger.info("Prediction done.")
    #
    # logger.info("Execute postprocessing.")
    # results = model_script.postprocess()
    # logger.info("Postprocessing done.")
    #
    # logger.info("Execute saving process.")
    # db_utils.write_annotation_to_db(request_form, results, logger)
    # logger.info("Saving process done.")

    # 5. In CML case, delete temporary files..
    # if request_form["deleteFiles"] == "True":
    #     trainer_name = request_form["trainerName"]
    #     logger.info("Deleting temporary CML files...")
    #     out_dir = Path(cml_dir).joinpath(
    #         PureWindowsPath(request_form["trainerOutputDirectory"])
    #     )
    #     os.remove(out_dir / trainer.model_weights_path)
    #     os.remove(out_dir / trainer.model_script_path)
    #     for f in model_script.DEPENDENCIES:
    #         os.remove(trainer_file_path.parent / f)
    #     trainer_fullname = trainer_name + ".trainer"
    #     os.remove(out_dir / trainer_fullname)
    #     logger.info("...done")

# except Exception as e:
# logger.error('Error:' + str(e))
#   status_utils.update_status(key, status_utils.JobStatus.ERROR)
# finally:
#    del results, ds_iter, ds_iter_pp, model, model_script, model_script_path, model_weight_path, spec



from flask import Blueprint, render_template
from nova_server.utils import status_utils

ui = Blueprint("ui", __name__)

@ui.route('/')
def index():
    jobs = status_utils.get_all_jobs()
    return render_template('ajax_template.html', title='Current Jobs', jobs=jobs)

@ui.route('/data')
def data():
    return {'data': status_utils.get_all_jobs()}



"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import datetime
import sys
import pkg_resources
import pathlib
from version import __version__

# Parse arguments
if "--nightly" in sys.argv:
    nightly = True
    sys.argv.remove("--nightly")
else:
    nightly = False

# Settings
project_name = "hcai-nova-server"
here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")

install_requires=[
    "hcai-datasets",
    "hcai-nova-utils",
    "Flask == 2.0.2",
    "imbalanced-learn==0.8.1",
    "waitress==2.0.0",
    "ffmpegio"

]

# Adjustment for nightly build
if nightly:
    project_name += "-nightly"
    datestring = datetime.datetime.now().strftime("%Y%m%d%H%M")
    curr_version = pkg_resources.parse_version(__version__)
    __version__ = f"{curr_version.base_version}.dev{datestring}"
    install_requires=[
        "hcai-datasets-nightly",
        "Flask == 2.0.2",
        "imbalanced-learn==0.8.1",
        "waitress==2.0.0",
        "ffmpegio"

    ]

# Setup
setup(
    name=project_name,
    version=__version__,
    description="!Alpha Version! - This repository contains code to setup a computational backend server for the nova annotation tool. You can use this backend to train models of your choosing or create explanations for existing models.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/sampleproject",
    author="Dominik Schiller",
    author_email="dominik.schiller@uni-a.de",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
    ],
    packages=find_packages(exclude=["*tests*"]),
    python_requires=">=3.6, <4",
    install_requires = install_requires,
    include_package_data=True,
    package_data={"": ["*.csv"]},
    entry_points = {
        'console_scripts': [
            'nova-server = nova_server.app:run_app',
        ]
    }
)



import ctypes
import threading

status_lock = threading.Lock()
ml_lock = threading.Lock()
jc_lock = threading.Lock()
job_counter = 0

THREADS = {}


def ml_thread_wrapper(func):
    """
    Executing the function in a mutex protected thread for asynchronous execution of long-running ml tasks.
    The thread waits with execution unit the status_lock is released. To do any initialization before the thread starts acquire the status lock before creating the thread.
    :param func:
    :return: The thread
    """

    def wrapper(*args, **kwargs):
        global job_counter

        def lock(*args, **kwargs):
            try:
                # Only one ml thread can be active at the same time
                ml_lock.acquire()
                func(*args, **kwargs)
            finally:
                ml_lock.release()

        jc_lock.acquire()
        job_id = str(job_counter)
        job_counter += 1
        jc_lock.release()

        t = BackendThread(target=lock, name=job_id, args=args, kwargs=kwargs)
        return t

    return wrapper


def status_thread_wrapper(func):
    def wrapper(*args, **kwargs):
        try:
            status_lock.acquire()
            return func(*args, **kwargs)
        finally:
            status_lock.release()

    return wrapper


class BackendThread(threading.Thread):
    def __init__(self, target, name, args, kwargs):
        threading.Thread.__init__(self, target=target, name=name, args=args, kwargs=kwargs)

    def get_id(self):
        if hasattr(self, '_thread_id'):
            return self._thread_id
        for id, thread in threading._active.items():
            if thread is self:
                return id

    def raise_exception(self):
        thread_id = self.get_id()
        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id,
                                                         ctypes.py_object(SystemExit))
        if res > 1:
            ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, 0)



import logging
import threading
import pathlib
from nova_server.utils import status_utils


LOGS = {}


def get_logfile_name_for_thread(log_key):
    return log_key + "-job_" + threading.current_thread().name


def get_log_path_for_thread(log_key):
    name = get_logfile_name_for_thread(log_key)
    # TODO: add  dynamic log path from config
    return pathlib.Path(__file__).parent.parent / "logs" / (name + ".log")


def init_logger(logger, log_key):
    print("Init logger" + str(threading.current_thread().name))
    try:
        log_path = get_log_path_for_thread(log_key)
        status_utils.set_log_path(log_key, log_path)
        handler = logging.FileHandler(log_path, "w")
        handler.setFormatter(logging.Formatter(fmt="%(asctime)s %(message)s", datefmt="%Y-%m-%d %H:%M:%S"))
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)
        LOGS[log_key] = logger
        return logger
    except Exception as e:
        print(
            "Logger for {} could not be initialized.".format(
                str(threading.current_thread().name)
            )
        )
        raise e


def get_logger_for_thread(log_key):
    logger = logging.getLogger(get_logfile_name_for_thread(log_key))
    if not logger.handlers:
        logger = init_logger(logger, log_key)
    return logger


def remove_log_from_dict(log_key):
    LOGS.pop(log_key)
